#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
Utilities for handling MatPlotLib figures, only available when the
``[plot]`` :ref:`extra <installation-extras>` is installed.

If `matplotlib`_ is not found at import time, this module will have only
a :py:data:`plot_enabled` attribute, which will be :py:obj:`False`. If
`matplotlib`_ is found, on the other hand, :py:data:`plot_enabled` will
be :py:obj:`True`, and all the other documented functions and attributes
of the module will be present.

.. py:data:: plot_enabled

   A boolean value indicating whether the ``[plot]``
   :ref:`extra <installation-extras>` has been installed. If
   :py:obj:`False`, the API will be severely limited.


.. include:: /link-defs.rst 
"""


import math
from contextlib import contextmanager
from io import BytesIO


__all__ = ['plot_enabled']


try:
    from matplotlib import pyplot as plt
except ImportError:
    from . import _display_missing_extra
    _display_missing_extra('plot', 'matplotlib')
    plot_enabled = False
else:
    __all__.extend(['figure_context', 'save_figure', 'set_figure_size'])
    plot_enabled = True


if plot_enabled:
    @contextmanager
    def figure_context(*args, **kwargs):
        """
        A context manager that automatically closes the figure that it
        opens.

        Inspired by https://github.com/matplotlib/matplotlib/issues/5218/#issue-110729876
        """
        fig = plt.figure(*args, **kwargs)
        yield fig
        plt.close(fig)


    def save_figure(fig, file=None, size=None, **kwargs):#dpi=300, format='png'
        """
        Save the figure as an image using
        :py:meth:`matplotlib.figure.Figure.savefig`.

        The main value of this method is that it automatically saves to
        memory via a :py:class:`~io.BytesIO` object if a file is not
        specified.

        Parameters
        ----------
        fig:
            The figure to save
        file: str, file-like, or None
            If not :py:obj:`None`, there is no return value. If
            :py:obj:`None`, :py:class:`~io.BytesIO` containing the image
            will be returned. The output will be rewound to the start in
            that case. The default is :py:obj:`None`.
        size: sequence[int]
            The size of the figure in inches, as a 2-element sequence
            ``(w, h)``. If either of the elements is :py:obj:`None`, the
            aspect ratio of the figure will be preserved. Even if only
            the width is specified, `size` must be a sequence.


        All other arguments are passed through directly to
        :py:meth:`~matplotlib.figure.Figure.savefig`. Some common
        options include:

        dpi :
            The resolution of the output image in dots-per-inch.
        format :
            The output format. ``'png'``, ``'svg'``, ``'pdf'`` have good
            support. If not supplied, the default is explicitly set to
            ``'png'``.
        frameon :
            Whether of nor the figure background should be rendered.
            Defaults to :py:obj:`True` if not supplied.
        transparent :
            Whether or not the axes background should be rendered as
            transparent. Defaults to the inverse value of `frameon` if
            not supplied.
        bbox_inches :
            The portion of the figure to save. If ``'tight'``, try to
            use the entire figure. If unset, defaults to ``'tight'``.
        pad_inches :
            The amount of padding to add around the figure when
            `bbox_inches` is ``'tight'``.
        """
        output = BytesIO() if file is None else file
        if size is not None:
            set_figure_size(fig, *size)
        kwargs.setdefault('format', 'png')
        kwargs.setdefault('bbox_inches', 'tight')
        kwargs.setdefault('frameon', fig.frameon)
        kwargs.setdefault('transparent', not kwargs.get('frameon'))
        fig.savefig(output, **kwargs)
        if file is None:
            output.seek(0)
        return output


    def set_figure_size(fig, w=None, h=None):
        """
        Set the size of a figure in inches, optionally preserving the
        aspect ratio.

        If either or the size arguments is :py:obj:`None`, it will be
        scaled to preserve the current aspect ratio. If both are
        :py:obj:`None`, the size is not set at all.
        """
        curr_w, curr_h = fig.get_size_inches()
        if w is None:
            if h is None:
                return
            w = h * curr_w / curr_h
        elif h is None:
            h = w * curr_h / curr_w
        fig.set_size_inches(w, h)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
Math utility functions that are otherwise uncategorized.
"""

import itertools
import math
import numpy


__all__ = [
        'ellipse', 'full_width_half_max', 'primes_up_to', 'first_primes',
        'count_divisors', 'round_sig',
]


def round_sig(x, n):
    """
    Round the number `x` to `n` significant figures.

    Based on https://stackoverflow.com/a/3411435/2988730
    """
    if not x:
        return x
    digits = -math.floor(math.log10(abs(x))) + (n - 1)
    return round(x, digits)


def ellipse(*args, num_points=1e3, **kwargs):
    r"""
    .. py:function:: ellipse(a, [b=0,] c, d, e, f, *, numPoints=1e3)
    .. py:function:: ellipse(a, b, h, k, *, theta=0, numPoints=1e3)

    Return x-y arrays for an ellipse in one of two standard forms.

    The equations are either the quadratic form,

    .. math::

       ax^2 + bxy + cy^2 + dx + ey + f = 0

    or the simplified form,

    .. math::

       \frac{(x - h)^2}{a^2} + \frac{(y - k)^2}{b^2} = 1

    The first form allows for rotated ellipses through the :math:`bxy`
    coupling term. The second form allows it through the explicit
    (optional) angle parameter `theta`, specified in radians
    counterclockwise about ``(h, k)``.

    The number of points is specified by `numPoints`. Points are evenly
    distributed by angle, not by arc-length (unless the ellipse is a circle).
    The default number of points is 1000.

    Return
    ======
    x : numpy.ndarray
        Array of `num_points` x-coordinates.
    y : numpy.ndarray
        Array of `num_points` y-coordinates.

    Notes
    =====
    This code is loosely based on my Stack Overflow answer:
    http://stackoverflow.com/a/41231100/2988730, which is in turn
    loosely based on the forum post at
    http://www.sosmath.com/CBB/viewtopic.php?t=17029
    """
    t = numpy.linspace(0, 2 * math.pi, num_points)

    coeffCount = len(args)
    if 5 <= coeffCount <= 6:
        if coeffCount == 5:
            a, c, d, e, f = args
            b = 0
        if len(args) == 6:
            a, b, c, d, e, f = args
        if kwargs:
            raise ValueError(
                'Only num_points can be a keyword for quadratic form of ellipse'
            )

        discriminant = b**2 - 4 * a * c
        if discriminant >= 0.0:
            name = 'parabloa' if discriminant == 0.0 else 'hyperbola'
            raise ValueError('Discriminant shows this to be '
                             'a {}, not an ellipse'.format(name))
        if b == 0:
            theta = 0
        else:
            theta = 0.5 * math.atan(b / (a - c))
            sin = math.sin(theta)
            cos = math.cos(theta)
            aa = a * cos**2 + b * sin * cos + c * sin**2
            cc = a * sin**2 - b * sin * cos + c * cos**2
            dd = d * cos + e * sin
            ee = -d * sin + e * cos
            a, b, c, d, e = aa, 0, cc, dd, ee
        ax_x = 1.0 / math.sqrt(a)
        ax_y = 1.0 / math.sqrt(c)
        scale = math.sqrt(d**2 / (4 * a) + e**2 / (4 * c) - f)
        h = -d / (2 * a)
        k = -e / (2 * c)
        xx = h + ax_x * scale * numpy.sin(t)
        yy = k + ax_y * scale * numpy.cos(t)
        if theta:
            x = xx * cos - yy * sin
            y = xx * sin + yy * cos
    elif coeffCount == 4:
        ax_x, ax_y, h, k = args
        xx = ax_x * numpy.sin(t)
        yy = ax_y * numpy.cos(t)
        theta = float(kwargs.pop('theta', 0.0))
        if theta:
            sin = math.sin(theta)
            cos = math.cos(theta)
            x = h + xx * cos - yy * sin
            y = k + xx * sin + yy * cos
        if kwargs:
           raise ValueError('Only num_points and theta can be a keyword '
                            'argument for simple form of ellipse')
    else:
        raise ValueError('Argument list must have 4 to 6 positional elements')

    return x, y


def full_width_half_max(x, y, factor=0.5, baseline=0.0, interp='linear', *,
                        return_points=False):
    """
    Compute the full-width, half-max metric for a dataset.

    The full-width half-max is the distance between where the data rises
    to half of the maximum for the last time before the max itself and
    where it falls below half of the maximum for the first time above
    the maximum.

    The actual fraction of the maximum that is used can be adjusted with
    `factor` (which defaults to 0.5). `factor` can be a number or the
    string ``'sigma'``, in which case it will be set to
    :math:`e^{-\\frac{1}{4}}`, the height of a Gaussian with unit
    amplitude one standard devition away from the mean. A fixed non-zero
    baseline can also be provided so that it does not have to be
    subtracted from `y` before being passed in.

    Since the actual `y` array is unlikely to contain the exact half-max
    value, an interpolation can be done. Currently, the following
    interpolation methods are supported:

      - ``'linear'``: linear interpolation betweent the x- and y-values
        surrouding the actual half-max point.
      - ``'nearest'``: nearest neighbor, i.e., use the x-value of the
        y-value that is closest to the half-max.

    Normally, a single scalar is returned. If `return_points` is
    :py:obj:`True`, however, two two-element tuples are returned as the
    second and third argument. Each tuple will contain an x-y pair of the
    intersection coordinates used to approximate the main return value.
    The first tuple will be for the left (rising) edge and the second
    will be for the right (trailing) edge.
    """
    imax = numpy.argmax(y)
    if factor == 'sigma':
        factor = math.exp(-0.25)
    # Also can be written as factor * y[imax] + (1.0 - factor) * baseline
    halfmax = factor * (y[imax] - baseline) + baseline
    # Always use <= to allow first and last y-value if they match exactly.
    rising = (y[:imax] <= halfmax) & (y[1:imax + 1] > halfmax)
    falling = (y[imax:-1] > halfmax) & (y[imax + 1:] <= halfmax)
    if not numpy.any(rising):
        raise ValueError('left edge does not fall below {} of max'.format(factor))
    if not numpy.any(falling):
        raise ValueError('right edge does not fall below {} of max'.format(factor))
    # Select last rising, first falling index.
    # Initial [0] because of nonzero's weird tuple return
    rising_index = numpy.nonzero(rising)[0][-1] + numpy.arange(2)
    falling_index = numpy.nonzero(falling)[0][0] + imax + numpy.arange(2)
    if interp == 'linear':
        def linterp(x0, x1, y0, y1, y):
            return x0 + (x1 - x0) * (y - y0) / (y1 - y0)
        rising = (linterp(*x[rising_index], *y[rising_index], halfmax), halfmax)
        falling = (linterp(*x[falling_index], *y[falling_index], halfmax), halfmax)
    elif interp == 'nearest':
        def nearest(ys, inds, y):
            return inds[numpy.argmin(numpy.abs(ys[inds] - y))]
        rising_i = nearest(y, rising_index, halfmax)
        falling_i = nearest(y, falling_index, halfmax)
        rising = (x[rising_i], y[rising_i])
        falling = (x[falling_i], y[falling_i])
    else:
        raise ValueError('Invalid value of `interp` parameter. '
                         'Expected {"linear", "nearest"}, found "{}".'.format(interp))

    if return_points:
        return falling[0] - rising[0], rising, falling

    return falling[0] - rising[0]


def primes_up_to(n):
    """
    Generate a set containing all the primes less than or equal to `n`.

    `n` must be a number that represents an array size that can exist
    in memory. The implementation uses an extremely unoptimized version
    of the sieve of Eratosthenes.

    Parameters
    ----------
    n : int
        The largest number to generate primes up to (exclusive). If you want an
        inclusive range, add 1 to this input.

    Return
    ------
    primes : set
        A set of all the primes less than `n`.
    """
    sieve = numpy.arange(2, n, dtype=numpy.int_)
    for i in range(n - 2):
        p = sieve[i]
        if p:
            sieve[i+p::p] = 0
    return sieve[numpy.nonzero(sieve)]


def first_primes(n):
    """
    Generate a set with the first `n` prime numbers.

    This is a toy method that should probably not be used for large
    prime numbers. Instead of actively discarding all multiples of
    found primes, it checks new candidates against each element of the
    current set of primes.

    Parameters
    ----------
    n : int

    Return
    ------
    primes : set
    """
    primes = set()
    for i in itertools.count(2):
        if all(i % p for p in primes):
            primes.add(i)
        if len(primes) == n:
            break
    return primes


def count_divisors(n):
    """
    Counts the divisors of natural number `n`, including 1 and itself.

    For example, ``28`` has divisors ``1, 2, 4, 7, 14, 28``, so
    ``count_divisors(28) == 7``.
    """
    count = 0
    limit = int(math.sqrt(n))
    # Add one if the number is a square (limit won't be part of loop)
    extra = (limit**2 == n)
    for i in range(2, limit):
        if not n % i:
            count += 1
    # Add one to count because loop does not check if 1/n are factors
    return (count + 1) * 2 + extra



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
Functions for manipulating the structure of objects.

This includes things like spoofing module contents, copying classes and
functions, and automatically creating properties.
"""

from copy import copy
from functools import update_wrapper, WRAPPER_ASSIGNMENTS
import sys
from types import FunctionType, ModuleType


__all__ = [
    'HiddenPropMeta', 'update_module', 'copy_func', 'copy_class'
]


def update_module(current, other, recurse=False):
    """
    Updates the dict of the module `current` with the dict of
    `other`.

    Either input may be a string (full name as given by ``__name__``),
    or a module reference.

    All functions and classes in `other` whose module is `other`
    will by default be copied and reassigned to `current`. All other
    non-dunder attributes will be copied exactly. Dunder attributes will
    be skipped, except ``__all__``, which will be shallow-copied as a
    list.

    Any attributes already defined in `current` will be skipped. This
    ensures that the globals defined in new methods will be updated
    correctly for the new module.

    If `recurse` is set to :py:obj:`True`, any sub-modules of `other`
    will be copied using this method instead of referenced directly.
    """
    def getmod(mod):
        if isinstance(mod, str):
            return sys.modules[mod]
        # Ducktype non-string objects. This allows things like
        # classes spoofing a module to pass through just fine.
        return mod
    current = getmod(current)
    other = getmod(other)
    cn, on = current.__name__, other.__name__
    cd = current.__dict__
    for name, item in other.__dict__.items():
        if name in cd:
            continue
        if name.startswith('__') and name.endswith('__'):
            if name == '__all__':
                try:
                    item = item[:]
                except Exception:
                    pass
            else:
                continue
        elif isinstance(item, FunctionType):
            if item.__module__ == on:
                item = copy_func(item, cd, cn)
        elif isinstance(item, type):
            if item.__module__ == on:
                item = copy_class(item, cd, cn)
        elif isinstance(item, ModuleType) and recurse:
            if item.__name__.startswith(on + '.'):
                mod = ModuleType(item.__name__.replace(on, cn), item.__doc__)
                update_module(mod, item, recurse)
                sys.modules[mod.__name__] = mod
                item = mod
        cd[name] = item


def copy_class(c, globals=None, module=None):            
    """
    Creates a shallow copy of a class object, optionally replacing
    its module and the object its methods reference for globals.

    This function is useful when importing a class into another
    module, and having it behave as a class of the importing module::

        from mod import cls
        cls = copy_class(cls, globals(), __name__)

    Parameters
    ----------
    c : type
        The class to copy
    globals : dict or None
        If :py:obj:`None`, copy the global dictionaries referenced by
        the methods unchanged. A popular alternative is ``globals()``.
    module : str or None
        The name of the module that this class belongs to. If
        :py:obj:`None`, keep ``c.__module__`` and the modules of all
        methods directly. A popular alternative is ``__name__``.

    Notes
    -----
    This function may not work properly for classes whose metaclass does
    not invoke :py:meth:`type.new <object.__new__>` at some point in the
    construction process.

    Based on https://stackoverflow.com/a/49157516/2988730.
    """
    if module is None:
        module = c.__module__
    attrs = dict(c.__dict__)
    for name, item in c.__dict__.items():
        if isinstance(item, type):
            attrs[name] = copy_class(item, globals, module)
        elif isinstance(item, FunctionType):
            attrs[name] = copy_func(
                item, item.__globals__ if globals is None else globals, module
            )
    d = type.__new__(type(c), c.__name__, c.__bases__, attrs)
    d.__module__ = module
    return d


def copy_func(f, globals=None, module=None):
    """
    Creates a shallow copy of a function object, optionally replacing
    the object it references for its globals.

    This function is useful when importing a function into another
    module, and having it behave as a function of the importing module::

        from mod import func
        func = copy_func(func, globals(), __name__)

    Parameters
    ----------
    f : function
        The object to copy
    globals : dict or None
        If :py:obj:`None`, copy the global dictionary referenced by
        `f`. A popular alternative is ``globals()``.
    module : str or None
        The name of the module that this function belongs to. If
        :py:obj:`None`, copy ``f.__module__`` directly. A popular
        alternative is ``__name__``.

    Notes
    -----
    Based originally on https://stackoverflow.com/a/13503277/2988730,
    and updated in https://stackoverflow.com/a/49077211/2988730.
    """
    g = FunctionType(f.__code__, f.__globals__ if globals is None else globals,
                     name=f.__name__, argdefs=f.__defaults__,
                     closure=f.__closure__)
    g = update_wrapper(g, f)
    if module is not None:
        g.__module__ = module
    g.__kwdefaults__ = copy(f.__kwdefaults__)
    return g


class HiddenPropMeta(type):
    """
    Creates a class with "hidden" read-only properties named in the
    :py:attr:`__hidden_properties__` attribute.

    A hidden property is one that stores its value under a 
    :py:attr:`~object.__dict__` key with the same name. This meta-class
    is therefore incompatible with anything that uses
    :py:obj:`~object.__slots__`.

    .. py:attribute:: __hidden_properties__

       This can be a single string, an iterable of strings, or an
       iterable of two-element tuples containing a string name and an
       initial value. Strings and tuples may be mixed together in an
       iterable. The attribute will be removed from the class body by
       this metaclass after it is processed.

    If the class has an explicit ``__init__`` method defined, it will be
    properly decorated to set the default values of the hidden
    properties. If an explicit ``__init__`` is not found, the implicit
    ``super.__init__`` constructor will be decorated in the same way and
    set as the initializer.
    """
    def __new__(meta, name, bases, body, **kwargs):
        if '__hidden_properties__' in body:
            props = body['__hidden_properties__']
            del body['__hidden_properties__']
            if isinstance(props, str):
                props = [props]

            def check_prop(prop):
                """
                Check if `prop` is a string or a 2-element tuple and
                returns a 2-element tuple. The first element must be a
                string, the second defaults to :py:obj:`None`.
                """
                if isinstance(prop, str):
                    return (prop, None)
                if not isinstance(prop, tuple) or len(prop) != 2 or \
                   not isinstance(prop[0], str):
                    raise ValueError(
                        'property description must be name or name-value tuple'
                    )
                return prop

            props = dict(check_prop(prop) for prop in props)

            def fget(name):
                """
                Create the getter for a given property name.
                """
                def getter(self):
                    return self.__dict__[name]
                getter.__name__  = name
                getter.__doc__ = f'Retrieves {name} as a read-only property.'
                return getter

            for prop in props:
                if prop not in body:
                    body[prop] = property(fget(prop))
                elif not isinstance(body[prop], property):
                    raise ValueError(
                        'property {!r} shadows non-property '
                        'attribute'.format(prop)
                    )
        else:
            props = None

        cls = super().__new__(meta, name, bases, body)
        if props:
            def wrapper(parent, initializer):
                def  __init__(self, *args, **kwargs):
                    self.__dict__.update(initializer)
                    return parent(self, *args, **kwargs)

                update_wrapper(
                    __init__, parent, tuple(
                        x for x in WRAPPER_ASSIGNMENTS if x != '__qualname__'
                    )
                )
                __init__.__qualname__ = '.'.join(
                    (cls.__qualname__, '__init__')
                )
                return __init__
            setattr(cls, '__init__', wrapper(getattr(cls, '__init__'), props))
        return cls



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding



"""
Implementations based on some of the recipes provided in the Python
documentation, and other sources like Stack Overflow.
"""


__all__ = [
    'all_combinations', 'all_nsc', 'any_nsc', 'consume', 'dict_merge',
    'dict_select', 'grouper', 'immutable', 'islast', 'is_ordered_subset',
    'lenumerate', 'mapping_context', 'option_lookup', 'shift_left',
    'KeyedSingleton', 'RecursiveDict', 'Namespace',
]


from collections import deque
from collections.abc import Mapping
from functools import reduce
from itertools import chain, combinations, islice, repeat
from operator import index, or_, and_
from warnings import warn

from .exceptions import ErrorTransform


def immutable(allow_properties=True,
              docstring='Forbids write access to class attributes.'):
    """
    Create a class decorator that sets a pre-defined version of
    :py:meth:`~object.__setattr__` that forbids write access to the
    attributes of a class.

    Access to properties may be allowed through `allow_properties`.
    The default setting of :py:obj:`True` does not guarantee that
    properties will be acessible, just that the parent implementation
    of :py:meth:`~object.__setattr__` will be responsible for handling
    the access.

    The messages of exceptions that are raised are based on the type of
    access that is requested. They mimic the default message that appear
    from most Python classes.

    Parameters
    ----------
    allow_properties : bool
        Whether or not to make an exception for properties. If
        :py:obj:`True`, setting of properties will be delegated to the
        parent's :py:meth:`~object.__setattr__`.
    docstring : str
        The string to associate with :py:meth:`~object.__setattr__`\ 's
        `__doc__` attibute.

    Return
    ------
    decorator : callable
        A class decorator that inserts a :py:meth:`~object.__setattr__`
        function into the decorated class.

    Notes
    -----
    Checking for properties based on the following Stack Overflow post:
    https://stackoverflow.com/a/46101204/2988730. See also the
    discussion following the answer itself.

    MRO issues with properly calling :py:meth:`~object.__setattr__` on
    properties are discussed in great detail in the answers and
    discussion surrounding https://stackoverflow.com/q/46121637/2988730.
    """
    def make_setattr(owner):
        if allow_properties:
            def __setattr__(self, name, value):
                typ = type(self)
                prop = getattr(typ, name, None)
                if isinstance(prop, property):
                    if owner is None:
                        return prop.__set__(self, value)
                    return super(owner, self).__setattr__(name, value)

                if hasattr(self, name):
                    msg = 'readonly attribute'
                else:
                    msg = '{!r} object has no attribute {!r}'.format(
                        type(self).__name__, name
                    )
                raise AttributeError(msg)
        else:
            def __setattr__(self, name, value):
                if hasattr(self, name):
                    msg = 'readonly attribute'
                else:
                    msg = '{!r} object has no attribute {!r}'.format(
                        type(self).__name__, name
                    )
                raise AttributeError(msg)

        if docstring:
            __setattr__.__doc__ = docstring

        return __setattr__

    def decorator(cls):
        """
        A class decorator that assigns a special
        :py:meth:`~object.__setattr__` method to forbid attribute
        modification.
        """
        cls.__setattr__ = make_setattr(cls)
        return cls

    return decorator


def consume(iterator, n=None):
    """
    Consume an iterator entirely, or advance `n` steps ahead.

    This function is based pretty much exactly on the similarly named
    recipe in the documentation at
    https://docs.python.org/3/library/itertools.html#itertools-recipes
    """
    # Use functions that consume iterators at C speed.
    if n is None:
        # feed the entire iterator into a zero-length deque
        deque(iterator, maxlen=0)
    else:
        # advance to the empty slice starting at position n
        next(islice(iterator, n, n), None)


def dict_merge(parent, child, keys=None, exclude=None, key=None):
    """
    Filter a parent dictionary and override it with values from a
    child, if supplied as a mapping.

    Parameters
    ----------
    parent : dict
        The base mapping to get the keys from.
    child : dict, value or None
        If a mapping type, `child[key]` will be the main value and the
        remaining keys will override the values obtained from `parent`.
        Otherwise, `child` will be the main value. If :py:obj:`None`,
        treated as missing entirely.
    keys : iterable or None
        An iterable of the keys to extract from `parent` and `child`, if
        it is a mapping. Should not contain `key`. :py:obj:`None` means
        to use all the keys of `parent`.
    exclude : container or None
        Keys to omit from the final result, even if they are present in
        `keys`.
    key :
        The name of the main value in `parent`, if it is not overriden
        by `child`. If `child` is a mapping, either it or `parent` must
        contain this key. Otherwise, the value of child is the value of
        the key.

    Return
    ------
    main: value (opt)
        The value of the main key, only returned if `key` is not
        :py:obj:`None`. If `child` is a mapping type, either it or
        `parent` will provide the key. Otherwise, it will be `child`
        itself.
    selection:
        The selected values of `parent`, possibly overriden by `child`
        if it is a mapping.
    """
    selection = dict_select(parent, keys, exclude)
    cmap = isinstance(child, Mapping)

    if cmap:
        selection.update(dict_select(child, keys, exclude))

    if key is None:
        return selection

    if child is None:
        value = parent[key]
    elif cmap:
        value = child[key] if key in child else parent[key]
    else:
        value = child

    return value, selection


def dict_select(dic, keys=None, exclude=None, extra='ignore'):
    """
    Filter a dictionary so only the specified keys are present.

    Parameters
    ----------
    dic : dict
        The dictionarty to filter.
    keys : iterable or None
        The keys to include in the output. Another dictionary may be
        used since it iterates over its keys by default. :py:obj:`None`
        means to use all keys.
    exclude : container or None
        The keys to exclude. Anything that supports the `in` operator is
        valid here. `exclude` takes precedence over `keys`: no keys in
        `exclude` will be present in the output, even if they are
        present in `keys`.
    extra : {'ignore', 'err', 'warn'}
        How to handle members of `dic` that are neither in `keys` nor
        explicitly listed in `exclude`:

        ignore :
            Skip over extra keys.
        err :
            Issue an error if invalid keys are found.
        warn :
            Issue a warning if extra keys are found.

        Values other than `'ignore'` will compare sets of keys.

    Return
    ------
    selection : dict
        A new `dict` object, even if `keys` is a superset of the actual
        keys found in `dic`.

    Notes
    -----
    The default behavior is just to make a copy of `dic`.
    """
    input_keys = set(dic.keys())
    select_keys = input_keys if keys is None else set(keys)
    exclude_keys = set() if exclude is None else set(exclude)

    if extra != 'ignore':
        extra_keys = (input_keys - select_keys) - exclude_keys
        extra_fmt = ', '.join(str(k) for k in extra_keys)
        message = 'Found extra keys: {}'.format(extra_fmt)
        if extra == 'warn':
            warn(message)
        elif extra == 'err':
            raise KeyError(message)
        else:
            raise ValueError('Invalid value of `extra`: "{}"'.format(extra))

    selection = (input_keys & select_keys) - exclude_keys

    return {k: dic[k] for k in selection}


def grouper(iterable, n, tail=True, fillvalue=None):
    """
    Gather an iterable into chunks or blocks of fixed length.

    Based on ``grouper`` from :ref:`itertools-recipes`. The main
    difference is that this implementation allows the final chunk to
    contain fewer elements. This version also returns lazy iterables for
    the inner chunks, unlike the ``grouper`` recipe which returns
    chunks.

    By default, the iterator assumes that all of the elements of a group
    have been consumed by the caller. If this is not the case, the next
    group will start with the un-consumed segment. An easy way to
    guarantee proper consumption is to use :py:func:`consume`.

    Inspired by https://stackoverflow.com/a/23926929/2988730.

    Parameters
    ----------
    iterable : iterable
        The iterable to split.
    n : int
        The size of the chunks to split into.
    tail : bool
        If :py:obj:`True`, the final group may contain fewer than `n`
        elements. If :py:obj:`False`, `fillvalue` will be used to pad
        the last group if it turns out to be shorter than `n`.
    fillvalue :
        The item to pad the final chunk with if `tail` is
        :py:obj:`False` and the iterable's length is not a multiple of
        `n`.

    Yields
    ------
    group :
        An iterable containing the next `n` or fewer elements of
        `iterable`.

    Notes
    -----
    This uses the updated generator protocol as per
    https://stackoverflow.com/a/45605358/2988730, meaning that this
    function theoretically requires Python 3.5+.
    """
    iterable = iter(iterable)
    n = index(n)
    while True:
        try:
            peek = next(iterable)
        except StopIteration:
            return
        it = iterable if tail else chain(iterable, repeat(fillvalue))
        yield chain((peek,), islice(it, n - 1))


def islast(iterable):
    """
    An iterator similar to the builtin :py:func:`enumerate`, except that
    instead of the index, it returns a boolean flag indicating if the
    current element is the last one.

    Based on the recipe at http://stackoverflow.com/a/2429118/2988730.
    Saw http://stackoverflow.com/a/2429260/2988730 after writing this.
    """
    iterable = iter(iterable)
    item = next(iterable)
    for peek in iterable:
        yield False, item
        item = peek
    yield True, item


def lenumerate(iterable, seq_type=None):
    """
    A generator that returns the length of each element of `iterable` in
    addition to the element itself.

    Each element must be an iterable itself. `seq_type` controls the
    type of sequence the elements will be turned into to get the length.
    If `None`, elements are assumed to have a `len` and will not be
    altered. Otherwise, `seq_type` must be a callable that accepts one
    iterable argument and returns a valid sequence, like
    :py:class:`list`, :py:class:`tuple`, :py:class:`str` or
    :py:class:`bytearray`.

    Specifying `seq_type` is roughly equivalent to calling
    ``lenumerate(map(seq_type, iterable))``. It guarantees consumption
    of all the elements of the sub-iterables. This could be important
    for something like :py:func:`grouper`.

    The name is a portmanteau of :py:func:`len` and
    :py:func:`enumerate`.
    """
    if seq_type is None:
        seq_type = lambda x: x
    for item in iterable:
        item = seq_type(item)
        yield len(item), item


def is_ordered_subset(a, b):
    """
    Check if iterable `a` is an ordered subset of iterable `b`.

    An ordered subset means that elements can be selected from `b`
    while preserving order in such a way as to get `a`.

    For example,::

        0, 1, 2

    is an ordered subset of::

        0, 2, 1, 3, 2

    but::

        0, 3, 1

    is not.

    Source: https://stackoverflow.com/a/11820887/2988730
    """
    it = iter(b)
    return all(x in it for x in a)


def all_combinations(sequence, start=0, stop=None):
    """
    Generate all combinations of a sequence with all possible lengths.

    The smallest length is `start`, which defaults to zero. The largest
    is `stop`, which defaults to `len(sequence)`. `stop` is inclusive.

    To use an iterable with no :py:func:`len`, specify `stop` manually.

    Example::

        all_combinations([1, 2, 3])

    generates::

        (,), (1,), (2,), (3,), (1, 2), (1, 3), (2, 3), (1, 2, 3)
    """
    if stop is None:
        stop = len(sequence)
    for n in range(start, stop + 1):
        for combo in combinations(sequence, n):
            yield combo


class KeyedSingleton(type):
    """
    A meta-class for genralized singleton initialization that returns an
    existing object based on an input key.

    The key is expected to be the first argument to the class
    constructor.

    Inspired by https://stackoverflow.com/a/8665179/2988730 and
    https://stackoverflow.com/a/31723324/2988730.

    I added my own, improved answer at
    https://stackoverflow.com/a/45175660/2988730
    """
    def __init__(cls, *args, **kwargs):
        """
        Initializes the class with an empty dictionary of instances.

        Arguments are `name`, `bases`, and `attributes`, as usual.
        """
        super().__init__(*args, **kwargs)
        cls._instances = {}

    def __call__(cls, *args, **kwargs):
        """
        The constructor/initializer require at least one argument since
        the first argument is the singleton key.

        If an instance was already created with the requested key, it is
        returned without being re-allocated or re-initialized. 
        """
        if not args:
            raise ValueError(
                'Instance key must be first argument in constructor'
            )
        key = args[0]
        if key in cls._instances:
            instance = cls._instances[key]
        else:
            instance = super().__call__(*args, **kwargs)
            cls._instances[key] = instance
        return instance

    def reset(cls):
        """
        Resets/clears the class's registry so that new isntances will be
        constructed for further calls to `cls`.
        """
        cls._instances.clear()


class RecursiveDict(dict):
    """
    Mapping that allows recursive lookup of keys.

    Lookup can be controlled by specifying a key type and/or a value
    type. A value suitable for recursion that is not found as a key
    raises a :py:exc:`KeyError`, while a value of the wrong type raises
    a :py:exc:`TypeError`. Requesting a key of the wrong type raises a
    :py:exc:`TypeError` as well.

    Only lookup operations are different from :py:class:`dict`\ :
    :py:meth:`__getitem__`, :py:meth:`setdefault` and :py:meth:`get`.
    :py:meth:`~dict.pop` and :py:meth:`~dict.popitem` are not modified.
    Full control is provided through the unaltered setter operations.

    Here is an example of how :py:attr:`value_type` affects the lookup::

        d = RecursiveDict({'a': 1, 'b': 'a'})
        print(d.setdefault('d', 'c'))  # Prints "c"

    vs::

        d = RecursiveDict({'a': 1, 'b': 'a'}, value_type=int)
        print(d.setdefault('d', 'c'))  # Raises "KeyError: 'c'"

    .. py:attribute:: key_type

       A class object indicating the type that may be used as a key.
       :py:obj:`None` indicates "any". The default is :py:class:`str`.
       Values of this type indicate recursive lookup.

    .. py:attribute:: value_type

       A class object indicating the type that may be used as a value.
       :py:obj:`None`, the default, indicates "any" type.
    """
    __slots__ = ('key_type', 'value_type')

    def __init__(self, *args, key_type=str, value_type=None, **kwargs):
        """
        Construct a new mapping with the specified key and value types.

        All normal :py:class:`dict` constructor arguments are accepted.

        `key_type` and `value_type` may be classes, tuples of classes,
        or :py:obj:`None`. The latter case is equivalent to "anything
        goes". The dictionary will accept keys and values of the wrong
        type, but will raise an error on lookup, so it is best to leave
        either `key_type` or `value_type` as :py:obj:`None`.

        If both types are constrained and there is overlap between them,
        `key_type` always takes precedence: a value that can be a
        `key_type` is always looked up as a key. Only when the lookup
        fails is the value returned.

        If `key_type` is constrained, but `value_type` is not, lookup
        continues until a :py:exc:`KeyError` is raised or a
        non-\ `key_type` value is encountered. Similarly, if only
        `value_type` is constrained, lookup continues until a
        :py:exc:`KeyError` is raised or a value of the requested type
        is found.

        There is no constraint on either (both are `None`), recursion
        will stop only when a value is not present in the dictionary as
        a key.
        """
        self.key_type = key_type
        self.value_type = value_type
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        """
        Recursively fetch an item from this dictionary.
        """
        return self._conditional_get(key, iskey=True)[1]

    def setdefault(self, key, default=None):
        """
        Set the key to the specified value if not found, return the
        value for the key.

        The default value is set exactly as specified, but the return
        value is dereferenced.
        """
        if key in self:
            return self[key]
        self[key] = default
        return self._conditional_get(default)[1]

    def get(self, key, default=None):
        """
        Retrieve a fully dereferenced value for `key` if present, or
        `default` if not.

        `default` gets dereferenced if the key is not present.
        """
        if key in self:
            return self[key]
        return self._conditional_get(default)[1]

    def final_key(self, key):
        """
        In a recursive reference, retreive the final key that actually
        contains the value mapping.

        For example, given::

            x = RecursiveDict(key_type=str, value_type=int)
            x['a'] = 'b'
            x['b'] = 'c'
            x['c'] = 1

        The result of :py:meth:`final_key` on any of the keys defined
        above would be ``'c'``.

        If the final key does not contain a value, it will still be
        returned. This means that if the dictionary above were to have
        ``x['c'] = 'd'`` (a broken recursion), :py:meth:`final_key`
        would return ``'c'`` regardless.

        Passing in a missing key will raise a :py:exc:`KeyError` as
        usual.
        """
        return self._conditional_get(key, iskey=True, err=False)[0]

    def _conditional_get(self, obj, iskey=False, err=True):
        """
        Check if the object is a possible key, and correctly retrieve
        the dereferenced value for it.

        If the value is a key type, but is not a valid key, it must also
        be a valid value type to avoid raising a :py:exc:`KeyError`.

        Parameters
        ----------
        obj :
            The object to search for.
        iskey : bool
            :py:obj:`True` if `obj` is required to be a key on the
            first iteration. Otherwise it represents a value (which may
            be a recursive key).
        err : bool
            Whether or not to raise an error if the key is not found.
        """
        key = obj if iskey else None
        while self._check_key_or_value_type(obj, iskey=iskey):
            try:
                new = super().__getitem__(obj)
            except KeyError:
                if err and (iskey or
                            not self._check_value_type(obj, err=False)):
                    raise
                break
            else:
                key = obj
                obj = new
            iskey = False
        return key, obj

    def _check_key_type(self, key, err=True):
        """
        Verify that the selected object is a valid key.

        Return the status, or raise a :py:exc:`TypeError` if `err` is
        :py:obj:`True`.
        """
        if self.key_type is None or isinstance(key, self.key_type):
            return True
        if err:
            raise TypeError('{} not allowed for keys'.format(type(key)))
        return False

    def _check_value_type(self, value, err=True):
        """
        Verify that the selected object is a valid value.

        Return the status, or raise a :py:exc:`TypeError` if `err` is
        :py:obj:`True`.
        """
        if self.value_type is None or isinstance(value, self.value_type):
            return True
        if err:
            raise TypeError('{} not allowed in mapping'.format(type(value)))
        return False

    def _check_key_or_value_type(self, obj, iskey):
        """
        Check if the object is of a proper key or value type, depending
        on the iteration count.

        If `iskey` is :py:obj:`True`, the object must be a key. All
        subsequent iterations return :py:obj:`True` if the object is a
        key, :py:obj:`False` if it is a valid value, and raise an error
        if it is neither.
        """
        check = self._check_key_type(obj, err=iskey)
        if check:
            return True
        return not self._check_value_type(obj, err=True)


def shift_left(*iterables):
    """
    For a given sequence of iterables, return a sequence of iterables
    that has all the elements shifted as far left as possible.

    For example, given::

        a = [1]
        b = [2, 3]
        c = [4, 5, 6]
        d = [7, 8]

    ``a, b, c, d = map(list, shift_left(a, b, c, d))`` will rearrange
    the lists so that::

        a = [1, 3, 6]
        b = [2, 5]
        c = [4, 8]
        d = [7]

    The name ``shift_left`` comes from visulizing the inputs as columns::

        1 2 4 7         1 2 4 7
          3 5 8   ==>   3 5 8
            6           6

    The result is a list of lists.
    """
    # Get iterators
    iterables = [iter(iterable) for iterable in iterables]
    # Generate a list of tuples containing each row
    count = len(iterables)
    shifted = [[] for _ in range(count)]
    while count > 0:
        # Generate a row
        it = 0
        while it < count:
            try:
                shifted[it].append(next(iterables[it]))
            except StopIteration:
                # Remove expired iterable
                del iterables[it]
                count -= 1
            else:
                # Keep moving along row
                it += 1

    return shifted


def option_lookup(name, mapping, option, key_func=None, value_func=None,
                  err_type=ValueError, key_err=KeyError):
    """
    Perform a lookup of an option in a mapping whose keys represent
    the valid options.

    If an invalid option is selected, it an `err_type` error is raised
    (:py:exc:`ValueError` by default).

    Parameters
    ----------
    name : str
        The name of the mapping to display if an error occurs. If the
        option is a function argument, the name of the argument is a
        good name choice.
    mapping : mapping
        The mapping to perform a the lookup in.
    option : object
        The key to look up in mapping.
    key_func : callable or None
        The transformation to apply to `option` to obtain the actual
        key to look up. If :py:obj:`None` (the default), no
        transformation is done. A common choice for this argument is
        :py:meth:`str.casefold` or :py:meth:`str.lower`, for
        case-insensitive string keys.
    value_func : callable or None
        The transformation to apply to the value that is found. If
        :py:obj:`None` (the default), no transformation is done. If
        supplied, must be a callable that accepts three arguments:

        - ``option``
        - ``key``: ``option`` transformed by ``key_func``
        - ``value``: The raw value retreived from the dictionary.

    err_type : type
        The type of the :py:exc:`Exception` to raise. Default is
        :py:exc:`ValueError`.
    key_err : type or tuple[type]
        The type of error expected from a failed lookup in `mapping`.
        If multiple types are to be expected, a :py:class:`tuple` of
        types may be supplied. Defaults to :py:exc:`KeyError`.
    """
    key = option if key_func is None else key_func(option)
    with ErrorTransform(key_err, err_type, 'Invalid value for `{0}`: {1!r}',
                        name, option):
        value = mapping[key]
    if value_func is not None:
        value = value_func(option, key, value)
    return value


class mapping_context:
    """
    A context manager for temporarily modifying the keys of a
    mapping.

    The context manager has a :py:meth:`delete` method that allows
    values to be removed as well. The method accepts any number of
    attribute names to delete and returns the context manager, so it can
    be used in a :ref:`with` block directly. Missing key names are
    ignored.

    All mapping values are reset to their original values when the
    manager exits. Nesting multiple instances of this context manager
    has a cumulative effect.

    The context manager is reentrant: the manager can be entered and
    exited and modified before re-entry as many times as necessary.

    Sample usage::

        d = {'a': 1, 'b': 2}
        print(d)
        with mapping_context(d, b=0, c=3):
            print(d)
            with mapping_context(d, b=1).delete('a'):
                print(d)
            print(d)
        print(d)

    Will result in::

        {'a': 1, 'b': 2}
        {'a': 1, 'b': 0, 'c': 3}
        {'b': 1, 'c': 3}
        {'b': 0, 'c': 3, 'a': 1}
        {'b': 2, 'a': 1}

    Instances have three documented attributes:

    .. py:attribute:: mapping

       The :py:class:`dict`\ -like object that this context manager
       applies to. The entire interface to the mapping is through the
       :py:meth:`getfunc` and :py:meth:`setfunc` methods, so that a
       context manager for non-mapping types can be simulated by
       subclassing.

    .. py:attribute:: updates

       A mapping of the added keys to their prior values. This mapping
       is unordered.

    .. py:attribute:: sentinel

       A special marker object guaranteed not to exists in the original
       mapping, used to mark deleted keys.
    """
    __slots__ = ('mapping', 'updates', 'sentinel')

    def __init__(self, mapping, *args, **kwargs):
        """
        Initialize a new context manager to replace the
        specified elements of the given mapping.

        Positional arguments represent mappings that contain keys and
        values to replace or add, each one in a format that would be
        accepted by :py:meth:`dict.update`.

        Keyword arguments are individual keys to update.
        """
        self.sentinel = object()
        self._init(mapping, *args, **kwargs)

    def __enter__(self):
        """
        Enter the context manager by updating the requested
        elements, return the context manager itself.
        """
        for key, value in self.updates.items():
            self.updates[key] = self.getfunc(key)
            self.setfunc(key, value)
        return self

    def __exit__(self, *args, **kwargs):
        """
        Restore the elements of the mapping.
        """
        # Convert to list to avoid modification during iteration
        for item in self.updates.items():
            self.setfunc(*item)

    def delete(self, *keys):
        """
        Remove all the keys named in ``keys`` from the mapping, but keep
        a record of them to be restored on exit.

        Return the context manager itself.
        """
        for key in keys:
            self.updates[key] = self.sentinel
        return self

    def chain(self, *args, **kwargs):
        """
        Create a chained :py:class:`mapping_context` with the same
        mapping class and :py:attr:`sentinel` object as this mapping.

        This method allows the example in the class docs to be rewritten
        as::

            d = {'a': 1, 'b': 2}
            print(d)
            with mapping_context(d, b=0, c=3) as mc:
                print(d)
                with mc.chain(b=1).delete('a'):
                    print(d)
                print(d)
            print(d)

        Returns a new :py:class:`mapping_context`.
        """
        cls = type(self)
        dc = cls.__new__(cls)
        dc.sentinel = self.sentinel
        dc._init(self.mapping, *args, **kwargs)
        return dc

    def _init(self, mapping, *args, **kwargs):
        """
        Sets ``mapping`` and records the modified and added keys
        similarly to :py:meth:`dict.update`, except that it allows
        multiple positional arguments, which may be of mixed types.
        Arguments are added in the following order:

        1. ``args`` is processed in order. Each element is added in the
           order of its iterator.
        2. ``kwargs`` is added in the order of its iterator.

        Later duplicate keys silently override earlier ones.
        """
        self.mapping = mapping
        self.updates = {}
        for arg in args:
            self.updates.update(arg)
        self.updates.update(kwargs)

    def getfunc(self, key):
        """
        A customizable function to get a single element of the mapping.

        If `key` is not present, this function must return
        :py:attr:`sentinel`.

        The default implementation works for most Python builtin mapping
        types that support a :py:meth:`~dict.get` method.
        """
        return self.mapping.get(key, self.sentinel)

    def setfunc(self, key, value):
        """
        A customizable function to set a single element of the mapping.

        If value is :py:attr:`sentinel`, the element should be deleted.

        The default implementation works for most Python builtin mapping
        types that support :py:meth:`~object.__delitem__` and
        :py:meth:`~object.__setitem__` methods.
        """
        if value is self.sentinel:
            del self.mapping[key]
        else:
            self.mapping[key] = value


class _NamespaceOverrideContext(mapping_context):
    """
    An extension of :py:class:`mapping_context` that is tailored for
    :py:class:`Namespace`.

    This class is designed specifically for :meth:`Namespace.override`.
    """
    __slots__ = ()

    def __init__(self, namespace, **kwargs):
        """
        Initialize a new context manager to replace the
        specified elements of the given namespace.
        """
        super().__init__(namespace, **kwargs)

    def __enter__(self):
        """
        Enter the context manager and return the mapping rather than
        `self`.
        """
        return super().__enter__().mapping

    def getfunc(self, key):
        """
        Mapping getter customized for :py:class:`Namespace`
        instead of regular :py:class:`dict`.
        """
        return getattr(self.mapping, key, self.sentinel)

    def setfunc(self, key, value):
        """
        Mapping setter and deleter customized for
        :py:class:`Namespace` instead of regular
        :py:class:`dict`.
        """
        if value is self.sentinel:
            delattr(self.mapping, key)
        else:
            setattr(self.mapping, key, value)


class Namespace:
    """
    A simple namespace object.

    The class is mutable. It implements containment checks. It can be
    converted to a dictionary using :py:func:`vars`.

    This class is based heavily on :py:class:`argparse.Namespace`.
    """

    def __init__(self, **kwargs):
        """
        Create a new namespace with the specified named arguments.
        """
        self.__dict__.update(kwargs)

    def __contains__(self, name):
        """
        Check if the namespace contains `name`.
        """
        return name in self.__dict__

    def __eq__(self, other):
        """
        Check if this object is equal to another namespace.

        Two namespaces are considered equal if they are both subclasses
        of :py:class:`Namespace` and contain the same mappings.
        """
        if not isinstance(other, __class__):
            return NotImplemented
        return vars(self) == vars(other)

    def __repr__(self):
        """
        Returns a usable string representation of this namespace.
        """
        items = ', '.join('{}={!r}'.format(*item)
                          for item in vars(self).items())
        return '{}({})'.format(type(self).__name__, items)

    def override(self, **kwargs):
        """
        Returns a context manager that can be used to temporarily set
        attributes in this namespace.

        The context manager has a `delete` method that allows values
        to be removed as well. The method accepts any number of
        attribute names to delete and returns the context manager, so it
        can be used in a :ref:`with` block directly. Missing key names
        are ignored.

        The context manager is a modified version of
        :py:class:`mapping_context`, so it also has a
        :py:meth:`~mapping_context.chain` method.

        All values will be reset to their original values when the
        manager exits. All context managers returned by this method
        operate on the same object, so their effects are cumulative.

        Sample usage::

            n = Namespace(a=1, b=2)
            print(n)
            with n.override(b=0, c=3):
                print(n)
                with n.override(b=1).delete('a'):
                    print(n)
                print(n)
            print(n)

        Will result in::

            Namespace(a=1, b=2)
            Namespace(a=1, b=0, c=3)
            Namespace(b=1, c=3)
            Namespace(b=0, c=3, a=1)
            Namespace(b=2, a=1)
        """
        return _NamespaceOverrideContext(self, **kwargs)


def any_nsc(iterable):
    """
    A non-short-circuiting version of :py:func:`any`.

    Useful for situations where the side-effects of an iterator are
    useful, for example when the elements are produced by a callable in
    a generator that also logs the :py:obj:`True` elements.

    The default behavior ``any([]) == False`` is preserved.

    See https://stackoverflow.com/q/1790520/2988730 for source material.
    """
    return reduce(or_, (bool(x) for x in iterable), False)


def all_nsc(iterable):
    """
    A non-short-circuiting version of :py:func:`all`.

    Useful for situations where the side-effects of an iterator are
    useful, for example when the elements are produced by a callable in
    a generator that also logs the :py:obj:`False` elements.

    The default behavior ``all([]) == True`` is preserved.

    See https://stackoverflow.com/q/1790520/2988730 for source material.
    """
    return reduce(and_, (bool(x) for x in iterable), True)
